/*
# PID controller for temperature control. 
This code is passed from Unknown=>张磊(11级)=》王东诣(12级)=》陈力(12级)
My main contribution: 
1. rewrite PID control.	Results are stored in float. In PID, proportion is pre-multiplied by 150 to reflect a reasonable relation between P, I and D.
2. Add PID result to display, easier for PID parameter modification.
3. solve the problem of shaking in display. Removing unnecessary delay and codes.
4. Remove keyboard lock.
*/
#include <reg51.h>   
#include <intrins.h>                            
#include <absacc.h>
#include <string.h> 
#include <math.h>

#define   uchar  unsigned char
#define   uint   unsigned int
#define   COM8279   XBYTE [0x0A001]   //8279 control
#define   DATA8279  XBYTE [0x0A000]   //8279 data

sbit     FAN=P3^3;    //FAN
sbit     DQ=P3^4;     //18B20
sbit     CR=P3^5;     //HEATER

unsigned int   temper;             //current temp
unsigned int   temp;               
unsigned int   set_temper=300;      //set temp
unsigned int   p_2=3,p_1=0,p_0=0;  //display temp，ten,one,digit
unsigned int   high_time,low_time,count=0;//PWM 
unsigned int   keyval;	         //key value
unsigned int   begin_judge1=0,begin_judge2=1,time=0;   

struct PID { 
	float SetPoint; // set target
	unsigned int Proportion; // P const
	unsigned int Integral; // I const
	unsigned int Derivative; // D const
	float LastError; // Error[-1] 
	float PrevError; // Error[-2] 
	float SumError; // Sums of Errors 
}; 

struct PID spid; // PID Control Structure 
float rout; // PID Response (Output) 
float rin; // PID Feedback (Input) 

uchar code ditab[16]={0x00,0x01,0x01,0x02,0x03,0x03,0x04,0x04,0x05,0x06,0x06,0x07,0x08,0x08,0x09,0x09};  //hex to dec，for 18b20 digit "-"
uchar led[]={0x3F,0x06,0x5B,0x4F,0x66,0x6D,0x7D,0x07,0x7F,0x6F,0x77,0x7C,0x39,0x5E,0x79,0x71}; //LED value fom 0 o F
uchar data display[5]={0x00,0x00,0x00,0x00,0x00};     //display data             
uchar data temp_data[2]={0x00,0x00};   //store temp read

void delay(uint t);
void key();
void ledclr();

void get_temper(){
	temper=display[2]*100+display[1]*10+display[0];
	set_temper=p_2*100+p_1*10+p_0;   
}

void delay(uint t){
  	for (;t>0;t--);
}

/*Initialize PID Structure*/ 
void PIDInit (struct PID *pp) {
	memset ( pp,0,sizeof(struct PID)); 
} 

/*PID*/ 
int PIDCalc( struct PID *pp, unsigned int NextPoint ) 
{ 
	int dError,Error; 
	Error = pp->SetPoint - NextPoint; // P
	pp->SumError += Error; // I 
	dError = pp->LastError - pp->PrevError; // D
	pp->PrevError = pp->LastError; 
	pp->LastError = Error; 
	return (pp->Proportion *  Error *150	+ pp->Integral * pp->SumError 	+ pp->Derivative * dError); // P+I+D 
} 

/*temp compare*/ 
compare_temper(){ 
	spid.SetPoint = set_temper;
	rout = PIDCalc ( &spid,rin ); // Perform PID Interation 
	if (high_time<=100) {
		if(rout>0){
			high_time=(unsigned int)(rout/80); 
		}
		else{
			high_time=(unsigned int)(-rout/80); 
		}
	}
	else{
		high_time=100; 
	}
	low_time= (100-high_time); 

	if(set_temper>temper){ 
		if(set_temper-temper>10){  //10 for 1 degree 
			high_time=100; 
			low_time=0; 
		} 
	} 
	else if(set_temper<temper){ 
		high_time=0; 
		low_time=100; 
	} 
	//delay except fan
	if(set_temper>temper-20){
		delay(500000);
	}
} 

//T0 interrupt  control PWM volt flip 40us*100=4ms 
void serve_T0() interrupt 1 using 2{ 
	if(++count<=(high_time)) 
		CR=1; 
	else if(count<=100){ 
		CR=0; 
	} 
	else{
		count=0; 
	} 
	TH0=0xec; 
	TL0=0x78; 
} 

//key interrupt
void intsvr1(void) interrupt 0 using 1{  
	COM8279=0x40;  //read FIFO
	keyval=DATA8279;
	key();
}

//clear led
void ledclr(){
	COM8279 = 0xD3; //clear RAM&FIFO  1101 0011
	delay(100);
	COM8279=0x00; //8279键盘双键锁定,内编码扫描，左端入口,8字符显示
	COM8279=0x32; //时钟分频设置,分频数为18   0011 0010
	COM8279=0x70; //设置读显示命令            0111 0000
	COM8279=0x90; //设置写显示命令            1001 0000
}

void key()
{
	EX0=0;  
	spid.SumError=0;
	//read key, set temp
	if(keyval== 0x0C3)
	  begin_judge1=1;
	if(keyval==0x0D0){
		if(spid.Proportion<15){
	    	spid.Proportion++;
		}
	}
	if(keyval==0x0D8){
		if(spid.Proportion>0){
			spid.Proportion--;
		}
	}
	if(keyval==0x0D1){
		if(spid.Integral<15){
			spid.Integral++;
		}
	}
	if(keyval==0x0D9){
		if(spid.Integral>0){
			spid.Integral--;
		}
	}
	if(keyval==0x0D2){
		if(spid.Derivative<15){
			spid.Derivative++;
		}
	}
	if(keyval==0x0DA){
		if(spid.Derivative>0){
			spid.Derivative--;
		}
	}
	//digit +
	if(keyval==0x0C0){
		if(p_0==9){
			if(p_1==9){
				p_2++;
				p_0=0;
				p_1=0;
			}
			else{
				p_1++;
				p_0=0;
			}
	  	}
		else{
			p_0++;
		}
	}
	//digit -
	if(keyval==0x0C8){
		if(p_0==0){
			if(p_1==0){
				p_2--;
				p_1=9;
				p_0=9;
			}
			else{
				p_1--;
				p_0=9;
			}
		}
		else{
		  p_0--;
		}
	}
	//ones +
	if(keyval==0x0C1){
		if(p_1==9){
			p_2++;
			p_1=0;
		}
		else{
			p_1++;		
		}
	}
	//ones -
	if(keyval==0x0C9){		
		if(p_1==0){
			p_2--;
			p_1=9;			
		}
		else{
			p_1--;
		}
	}
	//tens +
	if(keyval==0x0C2){
		p_2++;
	} 
	//tens -
	if(keyval==0x0cA){
		p_2--;
	}


// <0 or >=80 degree => set to 30 degree
	if(!((0<=p_2)&&(p_2<8))){
		ledclr();
	
		DATA8279=0x40;
		DATA8279=0x40;
		DATA8279=0x40;
		DATA8279=0x40;
		DATA8279=0x40;
		DATA8279=0x40;
		DATA8279=0x40;
		DATA8279=0x40;		
		delay(65535);
		delay(65535);
		p_2=3;
		p_1=0;
		p_0=0;
	
		EX0=1;
		return;
	}

//show temp after set
	if(keyval==0x0C8||keyval==0x0C0||keyval==0x0C1||keyval==0x0C9||keyval==0x0C2||keyval==0x0CA){
    	ledclr();
 		DATA8279=led[p_2];
		DATA8279=led[p_1]+0x80;
		DATA8279=led[p_0];
		DATA8279=0x00;
		DATA8279=0x63;//`C
		DATA8279=led[12];
		delay(50000);
		ledclr();
    	
		EX0=1;
	}
	if(keyval==0x0C3||keyval==0x0CB){
		delay(20000);
		EX0=1;
	}	
	if(keyval==0x0D8||keyval==0x0D0||keyval==0x0D1||keyval==0x0D9||keyval==0x0D2||keyval==0x0DA){
	    ledclr();
	    DATA8279=0x73;  //P
	    DATA8279=0x06;  //I
	    DATA8279=0xde;  //D.
	    DATA8279=0x00;
	    DATA8279=0x00;
  	    DATA8279=led[spid.Proportion];
	    DATA8279=led[spid.Integral];
	    DATA8279=led[spid.Derivative];
	    
	    delay(20000);
        delay(65535);
	    EX0=1;
	}
	spid.SumError=0;
	EX0=1;
}
//FAN control
void TempControl()  {	
	if((display[2]*100+display[1]*10+display[0])>(p_2*100+p_1*10+p_0+20)){
		FAN=0;
		FAN=1;
	}
	else{
		FAN=0;
		time=0;
	}
}

//display
void disp(){
	char k;
	char u;
	uchar data rdisplay[4]={0x00,0x00,0x00,0x00}; 
	unsigned int drout;
	
	if(begin_judge2==0){
		//Heating H
		if((display[2]*100+display[1]*10+display[0])<(p_2*100+p_1*10+p_0)-10){
			DATA8279=0x76;  
		}
		//Fanning F
		else if((display[2]*100+display[1]*10+display[0])>(p_2*100+p_1*10+p_0+20)){
			DATA8279=led[15];
		}
		//Pid controlling P
		else if((display[2]*100+display[1]*10+display[0])<=(p_2*100+p_1*10+p_0)){
			DATA8279=0x73; 
		}
		//not heating or Fanning 0 
		else{
			DATA8279=led[0]; 
		}
		DATA8279=0x00;
	}
	
	//temperature
	for(k=2;k>=0;k--){
		u=led[display[k]];
		if(k==1){
			u=u+0x80;//add do to digit
		}
		DATA8279=u;		
	}
	if(rout<0){
		DATA8279=0x40;
		drout=(unsigned int)(-rout)/80;
	}
	else{
		DATA8279=0x00;
		drout=(unsigned int)rout/80;
	}
	if(drout<100){
		rdisplay[0]=drout/100;//(unsigned int)rout/500;
		rdisplay[3]=drout%100;
		rdisplay[1]=rdisplay[3]/10;      
		rdisplay[2]=rdisplay[3]%10;
		
		DATA8279=led[rdisplay[1]];
		DATA8279=led[rdisplay[2]];
	}
	else{
		DATA8279=0x76;
		DATA8279=0x76;
	}
	//show ℃
	if(begin_judge2==0)
	{
		//DATA8279=0x00;
		//DATA8279=0x63;
		//DATA8279=led[12];//显示℃
	}
	TempControl();
}

//DS18B20 reset
ow_reset(void){
	char presence=1;
	while(presence){
		while(presence){
			DQ=1;_nop_();_nop_();//从高拉倒低
			DQ=0;                                       
			delay(50);           //550 us
			DQ=1;                                        
			delay(6);             //66 us
			presence=DQ;         //presence=0 复位成功,继续下一步
		} 
		delay(45);             //延时500 us
		presence=~DQ;         
	}
	DQ=1;                   //拉高电平
}

//DS18B20 write function
void write_byte(uchar val){ 
	uchar i;
	for(i=8;i>0;i--){
		DQ=1;_nop_();_nop_();                   //从高拉倒低
		DQ=0;_nop_();_nop_();_nop_();_nop_();   //5 us
		DQ=val&0x01;                           //最低位移出
		delay(6);                               //66 us
		val=val/2;                             //右移1位
	}
	DQ=1;
	delay(1);
}

//DS18B20 read 1 byte
uchar read_byte(void){
	uchar i;
	uchar value=0;
	for(i=8;i>0;i--){
		DQ=1;_nop_();_nop_();
		value>>=1;
		DQ=0;_nop_();_nop_();_nop_();_nop_();         //4 us
		DQ=1;_nop_();_nop_();_nop_();_nop_();         //4 us
		if(DQ)value|=0x80;
		delay(6);                                     //66 us
	}
	DQ=1;
	return(value);
}

//read temp
read_temp(){
	ow_reset();                   //reset
	delay(200);
	write_byte(0xcc);             //发命令
	write_byte(0x44);             //发转换命令
	ow_reset(); 
	delay(1);
	write_byte(0xcc);             //发命令
	write_byte(0xbe);
	delay(00);	
	temp_data[0]=read_byte();     //读温度值的低字节
	temp_data[1]=read_byte();     //读温度值的高字节
	temp=temp_data[1];
	temp<<=8;                                               
	temp=temp|temp_data[0];       // 两字节合成一个整型变量。
	return temp;                 
}

//temp data processing
work_temp(uint tem){
	uchar n=0;
	// temp >0 or <0. if <0 temp=>comp(temp),sign=1
	if(tem>6348){
		tem=65536-tem;
		n=1;
	}           
	display[4]=tem&0x0f;           // select digit part
	display[0]=ditab[display[4]];  // store digit 
	display[4]=tem>>4;             // select middle 8 bits(int part)
	display[3]=display[4]/100;     // select hundred bit
	display[1]=display[4]%100;     // select last 2 bits
	display[2]=display[1]/10;      // select tens bit
	display[1]=display[1]%10; 
	rin=display[2]*100+display[1]*10+display[0];
	
	if(!display[3]){
		display[3]=0x0a;           //when top bit is 0, do not display
		if(!display[2])
		{
			display[2]=0x0a;         //second top bit not display
		}
	}
	if(n){display[3]=0x0b;}       //<0 temp show "-"
}


main(){
	int a[5],a_final,count_a,max,min;
	set_temper=300;      //set current temp
	p_2=3,p_1=0,p_0=0;  
	count=0; 
	begin_judge1=0;
	begin_judge2=1;
	TMOD=0x21; 
	TH0=0xec; 
	TL0=0x78;	
	SCON=0x50; 
	PCON=0x00; 
	TH1=0xfd; 
	TL1=0xfd; 
	PS=1; 
	EA=1; 
	EX1=0; 
	ET0=1; 
	ES=1; 
	TR0=1; 
	TR1=1; 
	high_time=0; 
	low_time=100; 
	PIDInit (&spid);  
	spid.SumError=0;
	spid.Proportion = 4; // Set PID Coefficients 
	spid.Integral = 8; 
	spid.Derivative = 5; 
	spid.SetPoint = set_temper; // Set PID Setpoint 
	spid.LastError=0;
	EA=1;
	EX0=1;                             //enable interrupt；
	IT0=1;                              //interrupt using downing edge
	
    while(1){
		ledclr();
		DATA8279=led[p_2];
        DATA8279=led[p_1]+0x80;
        DATA8279=led[p_0];
        DATA8279=0x63;
        DATA8279=led[12];
        DATA8279=led[spid.Proportion];
        DATA8279=led[spid.Integral];
		DATA8279=led[spid.Derivative];
		
		delay(2000);
		if(begin_judge1==1){
			begin_judge2=0;
			break;
		}
	}
	while(1){	
			//select average of middle 3 in 5 read_temp
			a[0]=read_temp();
			a[1]=read_temp();
			a[2]=read_temp();
			a[3]=read_temp();
			a[4]=read_temp();
			for(count_a=0,max=a[0];count_a<5;count_a++){
				if(max<a[count_a]){
					max=a[count_a];
				}
			}
			for(count_a=0,min=a[0];count_a<5;count_a++){
				if(min>a[count_a]){
					min=a[count_a];
				}
			}
		a_final=(a[0]+a[1]+a[2]+a[3]+a[4]-max-min)/3;

		work_temp(a_final);                        
		//delay(50000);
		get_temper(); 
		compare_temper(); 
		disp();  
	}
}